#include "Cell.h"

Cell mapCells[240];

void initCells(u8* mapArray)
{
    u8 i, j;
    u8 position = 0; //if the grid size goes over 240 we must use u16

    for (i = 0; i < SCREEN_TILES_X; ++i)
    {            
        for (j = 0; j < SCREEN_TILES_Y; ++j)
        {
            //255 is empty, 0-99 is a tile(wall), 100-199 is an item, 200-254 is an interactable
            if(mapArray[position] == 255)
            {
                mapCells[position].type = CELL_TYPE_EMPTY;
				mapCells[position].data.tile = emptyTile;
            }
            else if(mapArray[position] >= INTERACTABLE_OFFSET)
            {
                mapCells[position].type = CELL_TYPE_INTERACTABLE;

                //interactables start at 200 so subtract 200 and we get the number from the interactable graphic set
                mapCells[position].data.interactable = mapArray[position] - INTERACTABLE_OFFSET;
            }
            else if(mapArray[position] >= ITEM_OFFSET)
            {
                mapCells[position].type = CELL_TYPE_ITEM;
				
                //items start at 100 so subtract 100 and we get the number from the item graphic set
                mapCells[position].data.item = mapArray[position] - ITEM_OFFSET;
            }
            else
            {
                mapCells[position].type = CELL_TYPE_TILE;
                mapCells[position].data.tile = mapArray[position];
            }
            
            ++position;
        }
    }
}

void drawCell(u8 x, u8 y, u16 position)
{
    u8* mapMemory = cpct_getScreenPtr(CPCT_VMEM_START, xPositions[x], yPositions[y]);
    
    switch (mapCells[position].type)
    {
		//treat enemy as empty so we can draw background underneath them
		case CELL_TYPE_ENEMY:
        case CELL_TYPE_EMPTY:
            cpct_drawSprite(tileGraphicSet[emptyTile], mapMemory, TILE_WIDTH_BYTES, TILE_HEIGHT);
        break;

        case CELL_TYPE_TILE:
            if(mapCells[position].data.tile == TILE_TYPE_WALL_GROUND)
            {
				cpct_drawSprite(tileGraphicSet[mapCells[position].data.tile], mapMemory, TILE_WIDTH_BYTES, TILE_HEIGHT);
            }
			else if (mapCells[position].data.tile == TILE_TYPE_WALL_MIDDLE)
			{
				cpct_drawSprite(tileGraphicSet[mapCells[position].data.tile], mapMemory, TILE_WIDTH_BYTES, TILE_HEIGHT);
			}
			else if (mapCells[position].data.tile == TILE_TYPE_WALL_TOP)
			{
				cpct_drawSprite(tileGraphicSet[mapCells[position].data.tile], mapMemory, TILE_WIDTH_BYTES, TILE_HEIGHT);
			}
			else if (mapCells[position].data.tile == TILE_TYPE_WALL_MIDDLE_INFESTED)
			{
				cpct_drawSprite(tileGraphicSet[mapCells[position].data.tile], mapMemory, TILE_WIDTH_BYTES, TILE_HEIGHT);
			}
			else if (mapCells[position].data.tile == TILE_TYPE_EMPTY_MIDDLE_INFESTED)
			{
				cpct_drawSprite(tileGraphicSet[mapCells[position].data.tile], mapMemory, TILE_WIDTH_BYTES, TILE_HEIGHT);
			}
			else if (mapCells[position].data.tile == TILE_TYPE_BURNT_WALL)
			{
				//draw the "burnt" wall as a standard wall as we are only using this id so we can update the map on segment move
				cpct_drawSprite(tileGraphicSet[wallTile], mapMemory, TILE_WIDTH_BYTES, TILE_HEIGHT);
			}
			else if (mapCells[position].data.tile == TILE_TYPE_BLACK)
			{
				//dark blue square where the enemy is going to be teleporting to
				cpct_drawSolidBox(mapMemory, cpct_px2byteM0(0, 0), TILE_WIDTH_BYTES, TILE_HEIGHT);
			}
            else
            {
                //other tiles are drawn as blue boxes ie water
                cpct_drawSolidBox(mapMemory, cpct_px2byteM0(0, 0), TILE_WIDTH_BYTES, TILE_HEIGHT);
            }
        break;

        case CELL_TYPE_ITEM:
            //items are drawn from the item graphic set
            cpct_drawSprite(itemGraphicSet[mapCells[position].data.item], mapMemory, TILE_WIDTH_BYTES, TILE_HEIGHT);
        break;

		case CELL_TYPE_INTERACTABLE:
			//interactables are drawn from the interactable graphic set
			cpct_drawSprite(interactableGraphicSet[mapCells[position].data.interactable], mapMemory, TILE_WIDTH_BYTES, TILE_HEIGHT);
		break;
    }
}

void drawCells()
{
    //draw the content of the mapArray (map grid)
    u8 x = 0;
    u8 y = 0;
    u16 position = 0;

    //position is the element in the map tile array
    for (x = 0; x < SCREEN_TILES_X; ++x)
    {
        for (y = 0; y < SCREEN_TILES_Y; ++y)
        {
            drawCell(x, y, position);

            ++position;
        }
    }
}