#include "Game.h"

u8 running_ = FALSE;

void initGame()
{
    running_ = TRUE;

    //setup keys struct for user input
    initKeys();

    //reset game state values for a new game
    resetGameState();

	//display the loading screen to tell the user to wait/what to do
	loadingScreen();

	//setup the map
	initMap(gameTimer % 200);

    //setup initial player position and sizes
    initPlayer();

	//setup the hud
	initHud();

	//initilise the draw queue
	initDrawQueue();
}

void runGame()
{
    while(running_)
    {
        //read input from the player/user
        readInput();

        //run the game logic which will effect the display
        gameLogic();

		//wait for vsync
		cpct_waitVSYNC();

        //draw the game display on the screen
        drawGame();
    }
}

void gameLogic()
{
    //move the player character
    movePlayer();

	if (gameControls.shootIsDown)
	{
		//player shooting a weapon if they have one
		playerShoot();
	}
    else
    {
        playerEndShoot();
    }

	//death trap timers
	//logicDeathTraps();

	//run enemy logic
	logicEnemies();
    	
	//the boss logic is only run on the last level in the bottom left segment
	if (currentLevel == TOP_LEVEL && currentSegment == BOSS_SEGMENT)
	{
		//run logic for the boss
		logicBoss();
	}

	//run the logic on the player character
	logicPlayer();

	//weapon switching
	weaponSwitch();

	//check if the player has completed the level
	if (nextLevel)
	{
		//set running to false to transition to the next level of the game
		running_ = FALSE;
	}
}

void drawGame()
{
    //draw the map grid on the screen
    drawMap();	

    //draw the player on the screen
    drawPlayer();
	
	//draw the enemies on the screen from the map
	drawEnemies();

    //hud will only update if redraw is true
    drawHud();
}

void readInput()
{
    //up
    if(cpct_isKeyPressed(gameControls.up))
    {
        gameControls.upIsDown = TRUE;
    }
    else
    {
        gameControls.upIsDown = FALSE;
    }

    //down
    if(cpct_isKeyPressed(gameControls.down))
    {
        gameControls.downIsDown = TRUE;
    }
    else
    {
        gameControls.downIsDown = FALSE;
    }

    //left
    if(cpct_isKeyPressed(gameControls.left))
    {
        gameControls.leftIsDown = TRUE;
    }
    else
    {
        gameControls.leftIsDown = FALSE;
    }

    //right
    if(cpct_isKeyPressed(gameControls.right))
    {
        gameControls.rightIsDown = TRUE;
    }
    else
    {
        gameControls.rightIsDown = FALSE;
    }

    //shoot
    if(cpct_isKeyPressed(gameControls.shoot))
    {
        gameControls.shootIsDown = TRUE;
    }
    else
    {
        gameControls.shootIsDown = FALSE;
    }

    if(cpct_isKeyPressed(gameControls.exit))
    {
        if(exitDialog())
        {
            running_ = FALSE;
        }
        else
        {
			//redraw the boss in case the exit dialog is displayed over the top of it
			gameBoss.redraw = TRUE;

            //force a map redraw to cleanup after the dialog is closed
            mapRedraw = TRUE;
        }
    }

	if (cpct_isKeyPressed(gameControls.objective))
	{
		objectiveDialog();

		//redraw the boss in case the objective dialog is displayed over the top of it
		gameBoss.redraw = TRUE;

		//force a map redraw to cleanup after the dialog is closed
		mapRedraw = TRUE;
	}

	if (cpct_isKeyPressed(gameControls.weapon))
	{
		//only set the timer off when this is first pressed
		if (gameControls.weaponIsDown == FALSE)
		{
			weaponSwitchTimer = gameTimer;
		}		

		gameControls.weaponIsDown = TRUE;
		
	}
	else
	{
		gameControls.weaponIsDown = FALSE;
	}
}

void runGameSession()
{
	//setup game session values
	resetGameSessionState();

	do
	{
		//setup the game
		initGame();

		//start the game for the current level
		runGame();

		//change level
		++currentLevel;
		
		//change the palette for the next level if not in green mode
		if (currentPalette != GREEN_PALETTE)
		{
			choosePalette(currentLevel);
		}

		if (currentLevel == MIDDLE_LEVEL)
		{
			wallTile = TILE_TYPE_WALL_MIDDLE;
			emptyTile = TILE_TYPE_EMPTY_MIDDLE;
		}
		else
		{
			//this should be top level
			wallTile = TILE_TYPE_WALL_TOP;
			emptyTile = TILE_TYPE_EMPTY_TOP;
		}

		//check to see if the player has completed the last level or if the player is dead
		if (currentLevel == LEVEL_COUNT || player.lives == 0)
		{
			nextLevel = FALSE;
		}
	} 
	while (nextLevel);

	//revert to ground elvel palette once game is ended
	if (currentPalette != GREEN_PALETTE)
	{
		choosePalette(GROUND_LEVEL);
	}
}