#include "Maze.h"

#define DIRECTION_COUNT_ 4
#define NORTH_X_ 0
#define NORTH_Y_ 1
#define SOUTH_X_ 0
#define SOUTH_Y_ -1
#define EAST_X_ 1
#define EAST_Y_ 0
#define WEST_X_ -1
#define WEST_Y_ 0

u8 mazeGrid[SCREEN_TILES_Y*SCREEN_TILES_X];
#define MAPM(x,y) mazeGrid[x*SCREEN_TILES_Y + y]

void addToPointXY(u8* pointX, u8* pointY, u8 addX, u8 addY)
{
    *pointX += addX;
	*pointY += addY;
}

u8 countBranchesXY(u8 x, u8 y, i8* directions)
{
	u8 branchCount = 0;
	u8 i = 0;
	

	for (i = 0; i < DIRECTION_COUNT_*2; i+=2)
	{
		u8 newX = x;
		u8 newY = y;

		//this produces a very interesting sparse effect to the maze
		//addToPointXY(&x, &y, directions[i], directions[i + 1]);
		        
        addToPointXY(&newX, &newY, directions[i], directions[i + 1]);

		//add either 1 or 0 based on the value
		if (MAPM(newX, newY) == EMPTY)
		{
			branchCount += 1;
		}
	}

	return branchCount;
}

u8 isValidPathXY(u8 x, u8 y, i8* directions, u8 wallType)
{
    //make sure within dimensions
    if((x > 0) && (y > 0) && (x < SCREEN_TILES_X - 1) && (y < SCREEN_TILES_Y - 1))
    {
        if(MAPM(x, y) == wallType)
        {
            u8 branchCount = countBranchesXY(x, y, directions);

            if(branchCount <= 1)
            {
                return TRUE;
            }
        }
    }

    return FALSE;
	//return (x > 0) && (y > 0) && (x < SCREEN_TILES_X - 1) && (y < SCREEN_TILES_Y - 1)
	//	&& (MAPM(x, y) == 0) && (countBranchesXY(x, y, directions) <= 1);
}

void generateMaze(u8 startX, u8 startY, u8 wallType, u8 seed)
{
    //define the possible directions
    u8 branchPointLength = 1;

	u8 pointToTestX = 255;
	u8 pointToTestY = 255;

	u8 branchPointsX[SCREEN_TILES_Y*SCREEN_TILES_X];
	u8 branchPointsY[SCREEN_TILES_Y*SCREEN_TILES_X];

	i8 directionsToTest[DIRECTION_COUNT_*2] = { NORTH_X_, NORTH_Y_, SOUTH_X_, SOUTH_Y_, EAST_X_, EAST_Y_, WEST_X_, WEST_Y_ };

	i8 nextDirectionX = -5;
	i8 nextDirectionY = -5;
	u8 newPointFound = 0;
	u8 newBranchPointX = 0;
	u8 newBranchPointY = 0;

	//fill the maze grid with solid wall of the specified type
	cpct_memset(mazeGrid, wallType, SCREEN_TILES_Y*SCREEN_TILES_X);


	branchPointsX[0] = startX;
	branchPointsY[0] = startY;

	//set the initial position to start the maze
	MAPM(startX, startY) = 255;

    cpct_srand(seed);
	while (branchPointLength > 0)
	{
        //starts with all directions being available
		u8 directionsTested = 0;

        u8 i = cpct_rand() % branchPointLength;

		pointToTestX = branchPointsX[i];
		pointToTestY = branchPointsY[i];

		//shuffle directions
		shuffleXY(directionsToTest);
		newPointFound = 0;

		while (!newPointFound && (DIRECTION_COUNT_ - directionsTested) > 0)
		{            
			//grab the next direction
			nextDirectionX = directionsToTest[directionsTested*2];
			nextDirectionY = directionsToTest[directionsTested*2 + 1];
			
			//next direction
			++directionsTested;

			//pointAdd(&pointToTest, &direction, &newBranchPoint);
			newBranchPointX = pointToTestX;
			newBranchPointY = pointToTestY;

			addToPointXY(&newBranchPointX, &newBranchPointY, nextDirectionX, nextDirectionY);
			
			if(isValidPathXY(newBranchPointX, newBranchPointY, directionsToTest, wallType))
			{
				//add the point to the branh points working array
				branchPointsX[branchPointLength] = newBranchPointX;
				branchPointsY[branchPointLength++] = newBranchPointY;

				//set this point as space
				MAPM(newBranchPointX, newBranchPointY) = 255;

				newPointFound = 1;
			}            
		}

		if (!newPointFound)
		{
			//swap the current item to the end of the array effectively deleting it
			//move not required when the last element of the working array
			if (i != branchPointLength - 1)
			{
				branchPointsX[i] = branchPointsX[branchPointLength-1];
				branchPointsY[i] = branchPointsY[branchPointLength-1];
			}

			//adjusting the length for the working array
			--branchPointLength;
		}
	}
}

void shuffleXY(i8* directionsToTest)
{
	u8 i = 0;

	for (i = (DIRECTION_COUNT_ - 1) * 2; i > 0; i -= 2)
	{
		u8 n = (cpct_rand() % DIRECTION_COUNT_) * 2;
		i8 tempX = directionsToTest[i];
		i8 tempY = directionsToTest[i + 1];

		directionsToTest[i] = directionsToTest[n];
		directionsToTest[i + 1] = directionsToTest[n + 1];

		directionsToTest[n] = tempX;
		directionsToTest[n + 1] = tempY;
	}
}

u8 addExit(u8 direction, u8 wallType)
{
	u8 startPosition = 0;

	if(direction == DIRECTION_EAST || DIRECTION_WEST)
	{
		//starts from the bottom left of the maze
		startPosition = SCREEN_TILES_Y - 1;
	}
	else
	{
		//starts from the bottom right of the maze
		startPosition = SCREEN_TILES_X - 1;
	}

	if(direction == DIRECTION_WEST)
	{
		//checks for a blank space one position left from the right edge 
		while (MAPM((SCREEN_TILES_X - 2), startPosition) == wallType && startPosition >= 0)
		{
			--startPosition;
		}

		//assuming there is a blank space one position from the right edge it adds a blank space to the right edge as the exit 
		if (startPosition >= 0) 
		{
			MAPM((SCREEN_TILES_X - 1), startPosition) = 255;
		}
	}
	else if(direction == DIRECTION_SOUTH)
	{
		//checks for a blank space one position above from the bottom edge 
		while (MAPM(startPosition, (SCREEN_TILES_Y - 2)) == wallType && startPosition >= 0)
		{
			--startPosition;
		}

		//assuming there is a blank space one position from the bottom edge it adds a blank space to the bottom edge as the exit 
		if (startPosition >= 0)
		{
			MAPM(startPosition, (SCREEN_TILES_Y - 1)) = 255;
		}
	}
	else if(direction == DIRECTION_EAST)
	{
		//checks for a blank space one position right from the left edge 
		while (MAPM(1, startPosition) == wallType && startPosition >= 0)
		{
			--startPosition;
		}

		//assuming there is a blank space one position from the left edge it adds a blank space to the left edge as the exit 
		if (startPosition >= 0)
		{
			MAPM(0, startPosition) = 255;
		}
	}
	else
	{
		//checks for a blank space one position below from the top edge 
		while (MAPM(startPosition, 1) == wallType && startPosition >= 0)
		{
			--startPosition;
		}

		//assuming there is a blank space one position from the top edge it adds a blank space to the top edge as the exit 
		if (startPosition >= 0)
		{
			MAPM(startPosition, 0) = 255;
		}
	}

	//return the position of the gap in the wall
	return startPosition;
}