#include "Player.h"
#include <stdio.h>

#define PLAYER_DEAD_LEVEL_ 7
#define BOSS_CELL_FOUND_ 3
#define WEAPON_SWITCH_TIME_ 255

//todo tune with difficulty
#define TEST_TUBE_TIME_ 3000

Player player;

u8 bulletBackgroundMemory[BULLET_PROJECTILE_WIDTH*BULLET_PROJECTILE_HEIGHT];
u8 laserBackgroundMemory_[BULLET_PROJECTILE_WIDTH*144];
u8 flameBackgroundMemory_[FLAME_PRJECTILE_HORIZONTAL_WIDTH*FLAME_PRJECTILE_HORIZONTAL_HEIGHT];

//used for ensuring objective dialog is only ever displayed once
u8 objectiveDialogFlag_;

u16 weaponSwitchTimer = 0;

void initPlayer()
{
	u8 x = 0;
	u8 y = 0;

    player.object.height = PLAYER_HEIGHT;
    player.object.width = PLAYER_WIDTH;

	//find a free place for the enemy
	determinePlayerPosition(&x, &y);
    player.object.position.x = x * TILE_WIDTH_BYTES;
	player.object.position.y = y * TILE_HEIGHT;

	playerEntryX = player.object.position.x;
	playerEntryY = player.object.position.y;

    //initially set previous position to 0 to initialise it to something different to position x,y
    player.object.previousPosition.x = 0;
    player.object.previousPosition.y = 0;

    //stationary
    player.object.velocity.x = 0;
    player.object.velocity.y = 0;

    //set player direction to north
    player.direction = DIRECTION_NORTH;
	
	//set starting life count for the player
    player.lives = livesCountStart;

    //set starting health for the player
    player.health = startingHealth;

    //player starts without a weapon and has no ammo
	player.currentWeapon = 0;
    player.hasProjectileLauncher = FALSE;
    player.projectileAmmo = 0;
    player.hasFlamethrower = FALSE;
    player.flamethrowerAmmo = 0;
    player.hasLaser = FALSE;
    player.laserAmmo = 0;

    //set inventory to empty
    player.inventory = 0x00;

	//the cell at the player position must be drawn before we can copy it
	drawCell(x, y, XY_TO_POSITION(x, y));

    //copy the content of the background where the player sprite will be drawn to save for later
    copyBackground(&player);

	//defaults to false
	objectiveDialogFlag_ = FALSE;

	//initialise to false for collision stuff
	player.collisionIndicator = FALSE;
	player.collisionFlickerCount = 0;
}

void drawPlayer()
{
	u8* memory;

	//draw the saved memory for where the projectile was previous (projectile.previousPosition)
	if (player.projectile.type == BULLET_PROJECTILE)
	{		
		drawPreviousBackgroundMemory(&player.projectile);
	}
	else if (player.projectile.type == FLAME_PROJECTILE)
	{
		if (player.projectile.object.previousPosition.x != player.projectile.object.position.x ||
			player.projectile.object.previousPosition.y != player.projectile.object.position.y)
		{
			drawPreviousBackgroundMemory(&player.projectile);
		}
	}

    //draw the player only when the position has changed
    if(player.object.previousPosition.x != player.object.position.x ||
       player.object.previousPosition.y != player.object.position.y)
    {        
		memory = cpct_getScreenPtr(CPCT_VMEM_START, player.object.previousPosition.x, player.object.previousPosition.y);	

		//previous position will be -1 (255) when a segment changes and we don't want to draw the background sprite then
		if (player.object.previousPosition.x != 255 && player.object.previousPosition.y != 255)
		{
			cpct_drawSprite(player.backgroundSprite, memory, TILE_WIDTH_BYTES, TILE_HEIGHT);
		}

		//todo: it might be better to copy only part of the sprite instead of everything but this is good initially
		

		//copy the content of the background where the player sprite will be drawn to save for later
        copyBackground(&player);
    }

	//copy the content of the background where the projectile will be drawn to save for later
	if (player.projectile.type == BULLET_PROJECTILE)
	{
		copyProjectileBackground(&player.projectile);
	}
	else if (player.projectile.type == FLAME_PROJECTILE)
	{
		if (player.projectile.object.previousPosition.x != player.projectile.object.position.x ||
			player.projectile.object.previousPosition.y != player.projectile.object.position.y)
		{
			copyProjectileBackground(&player.projectile);
		}
	}

    //draw projectile for the player before the player so it doesnt copy the player from the background
    drawProjectile(&player.projectile);

	//draw the player test sprite
    memory = cpct_getScreenPtr(CPCT_VMEM_START, player.object.position.x, player.object.position.y);

	if (player.collisionIndicator == FALSE)
	{
		if (player.direction == DIRECTION_NORTH)
		{
			cpct_drawSprite(G_player_sprite_north, memory, player.object.width, player.object.height);
		}
		else if (player.direction == DIRECTION_EAST)
		{			
			cpct_drawSprite(G_player_sprite_east, memory, player.object.width, player.object.height);
		}
		else if (player.direction == DIRECTION_SOUTH)
		{
			cpct_drawSprite(G_player_sprite_south, memory, player.object.width, player.object.height);
		}
		else
		{
			//west
			cpct_drawSprite(G_player_sprite_west, memory, player.object.width, player.object.height);
		}
	}
	else
	{
		cpct_drawSolidBox(memory, cpct_px2byteM0(4, 4), player.object.width, player.object.height);
	}
}

void movePlayer()
{
	//player cannot move while shooting a laser
	if (gameControls.shootIsDown && player.projectile.type == LASER_PROJECTILE)
	{
		return;
	}
	else if (gameControls.shootIsDown && player.projectile.type == FLAME_PROJECTILE)
	{
		//vertical projectile means the projectile can go east or west
		if (player.direction == DIRECTION_NORTH || player.direction == DIRECTION_SOUTH)
		{
			if (gameControls.leftIsDown)
			{
				player.projectile.object.velocity.x = -1;
			}

			if (gameControls.rightIsDown)
			{
				player.projectile.object.velocity.x = 1;
			}
		}
		else
		{
			//horizontal projectile means the projectile can go north or south
			if (gameControls.upIsDown)
			{
				player.projectile.object.velocity.y = -1;
			}

			if (gameControls.downIsDown)
			{
				player.projectile.object.velocity.y = 1;
			}
		}

		//player cannot move while using flamethrower so return
		return;
	}

    //updates player velocity based on keyboard input
    if (gameControls.upIsDown)
	{
		player.object.velocity.y = -1;
	}

	if (gameControls.downIsDown)
	{
		player.object.velocity.y = 1;
	}

	if (gameControls.leftIsDown)
	{
		player.object.velocity.x = -1;
	}

	if (gameControls.rightIsDown)
	{
		player.object.velocity.x = 1;
	}
}

void logicPlayer()
{
	u8 enemyTest1 = 0;
	u8 enemyTest2 = 0;

    //the tiles that the player currently occupies
	u8 occupiedXTileLeft = 0;
	u8 occupiedXTileRight = 0;
	u8 occupiedYTileTop = 0;
	u8 occupiedYTileBottom = 0;

	//the position the player occupies following a change in velocity
	u8 futureXTileLeft = 0;
	u8 futureXTileRight = 0;
	u8 futureYTileTop = 0;
	u8 futureYTileBottom = 0;

	//reset to false
	objectiveDialogFlag_ = FALSE;

    //update the previous position before updating the current position with the velocity
    player.object.previousPosition.x = player.object.position.x;
    player.object.previousPosition.y = player.object.position.y;

    occupiedXTileLeft = (player.object.position.x) / TILE_WIDTH_BYTES;
	occupiedXTileRight = (player.object.position.x + player.object.width - 1) / TILE_WIDTH_BYTES;
    occupiedYTileBottom = (player.object.position.y + player.object.height - 1) / TILE_HEIGHT;
    occupiedYTileTop = (player.object.position.y) / TILE_HEIGHT;
 
	futureXTileLeft = (player.object.position.x + player.object.velocity.x) / TILE_WIDTH_BYTES;
	futureXTileRight = (player.object.position.x + player.object.width - 1 + player.object.velocity.x) / TILE_WIDTH_BYTES;
	futureYTileBottom = (player.object.position.y + player.object.height - 1 + player.object.velocity.y) / TILE_HEIGHT;
	futureYTileTop = (player.object.position.y + player.object.velocity.y) / TILE_HEIGHT;

    if(player.object.velocity.y > 0)
    {
        //move down
        if(player.object.position.y + player.object.velocity.y <= SCREEN_Y_NO_HUD - player.object.height)
        {
            if (MAPA(occupiedXTileLeft, futureYTileBottom).type == CELL_TYPE_EMPTY && MAPA(occupiedXTileRight, futureYTileBottom).type == CELL_TYPE_EMPTY)
			{
                player.object.position.y += player.object.velocity.y;
                player.direction = DIRECTION_SOUTH;
			}
            else
            {
				enemyTest1 = processCell(occupiedXTileLeft, futureYTileBottom);
				enemyTest2 = processCell(occupiedXTileRight, futureYTileBottom);

				//enemy not found so continue as if empty
				if (enemyTest1 == 1 && enemyTest2 == 1)
				{
					player.object.position.y += player.object.velocity.y;
					player.direction = DIRECTION_SOUTH;
				}
            }
        }
		else
		{
			//attempt to move segment
			moveSegment(DIRECTION_SOUTH, &player.object);
		}
    }
    else if(player.object.velocity.y < 0)
    {
        //move up
        if(player.object.position.y + player.object.velocity.y >= 0)
        {
            if (MAPA(occupiedXTileLeft, futureYTileTop).type == CELL_TYPE_EMPTY && MAPA(occupiedXTileRight, futureYTileTop).type == CELL_TYPE_EMPTY)
			{
                player.object.position.y += player.object.velocity.y;
                player.direction = DIRECTION_NORTH;
            }
            else
            {
				enemyTest1 = processCell(occupiedXTileLeft, futureYTileTop);
				enemyTest2 = processCell(occupiedXTileRight, futureYTileTop);

				//enemy not found so continue as if empty
				if (enemyTest1 == 1 && enemyTest2 == 1)
				{
					player.object.position.y += player.object.velocity.y;
					player.direction = DIRECTION_NORTH;
				}
            }
        }
		else
		{
			//attempt to move segment
			moveSegment(DIRECTION_NORTH, &player.object);
		}
    }

    if(player.object.velocity.x > 0)
    {
        //move right
        if(player.object.position.x + player.object.velocity.x <= SCREEN_X - player.object.width)
        {
            if (MAPA(futureXTileRight, occupiedYTileTop).type == CELL_TYPE_EMPTY && MAPA(futureXTileRight, occupiedYTileBottom).type == CELL_TYPE_EMPTY)
			{
                player.object.position.x += player.object.velocity.x;
                player.direction = DIRECTION_EAST;         
            }
            else
            {
				enemyTest1 = processCell(futureXTileRight, occupiedYTileTop);
				enemyTest2 = processCell(futureXTileRight, occupiedYTileBottom);

				//enemy not found so continue as if empty
				if (enemyTest1 == 1 && enemyTest2 == 1)
				{
					player.object.position.x += player.object.velocity.x;
					player.direction = DIRECTION_EAST;
				}
            }
        }
        else
        {
            //attempt to move segment
			moveSegment(DIRECTION_EAST, &player.object);
        }
    }
    else if(player.object.velocity.x < 0)
    {
        //move left
        if(player.object.position.x + player.object.velocity.x >= 0)
        {
            if (MAPA(futureXTileLeft, occupiedYTileTop).type == CELL_TYPE_EMPTY && MAPA(futureXTileLeft, occupiedYTileBottom).type == CELL_TYPE_EMPTY)
            {
                player.object.position.x += player.object.velocity.x;
                player.direction = DIRECTION_WEST;
            }
            else
            {
				enemyTest1 = processCell(futureXTileLeft, occupiedYTileTop);
				enemyTest2 = processCell(futureXTileLeft, occupiedYTileBottom);

				//enemy not found so continue as if empty
				if (enemyTest1 == 1 && enemyTest2 == 1)
				{
					player.object.position.x += player.object.velocity.x;
					player.direction = DIRECTION_WEST;
				}
            }
        }
        else
        {
            //attempt to move segment
			moveSegment(DIRECTION_WEST, &player.object);
        }
    }

	if (player.inventory & INVENTORY_ITEM_TEST_TUBE)
	{
		//check if timer has elapsed
		if (player.timer < gameTimer)
		{
			player.inventory &= ~INVENTORY_ITEM_TEST_TUBE;

			//redraw the hud as the test tube is used up
			redrawHud = TRUE;
		}
	}

	if (player.collisionFlickerCount && player.timer < gameTimer)
	{
		player.collisionIndicator = !player.collisionIndicator;
		--player.collisionFlickerCount;
		player.timer = gameTimer + 50;
	}

    //reset velocity to 0
    player.object.velocity.x = 0;
    player.object.velocity.y = 0;

    //run the logic for the projectile
    logicProjectile(&player.projectile, &player.object);

	//player loses all health when they colide with the boss
	if (enemyTest1 == BOSS_CELL_FOUND_ || enemyTest2 == BOSS_CELL_FOUND_)
	{
		player.health = 0;
	}

	//player on 0 health
	if (!player.health)
	{
		if (player.lives)
		{
			//player loses a life
			--player.lives;

			if (player.lives)
			{
				//todo what happens when a player loses a life but isn't dead
				player.health = startingHealth;

				//redestribute enemies in the level
				if (!(currentLevel == TOP_LEVEL && currentSegment == BOSS_SEGMENT))
				{
					//kill all the enemies on the map
					u8 i = 0;

					for (i = 0; i < segmentEnemies[currentSegment]; ++i)
					{

						//only need to destroy it if its already been destroyed
						if (enemies[i].destroyed == FALSE)
						{
							killEnemy(&enemies[i]);
						}
					}

					//redistribute the enemies into the new segment
					distributeEnemies();
				}

				player.object.position.x = playerEntryX;
				player.object.position.y = playerEntryY;

				//redraw the updated hud values
				redrawHud = TRUE;
			}
			else
			{
				//todo player dead now
				nextLevel = TRUE;
				
				//increase current level to above 3 so that it can be treated as game over
				currentLevel = PLAYER_DEAD_LEVEL_;
			}
		}
	}
}

u8 processCell(u8 x, u8 y)
{
    Cell* cell = &MAPA(x,y);

    //prevent the player from picking up items/using interactables while using the flamethrower
	if (player.projectile.type == FLAMETHROWER_WEAPON)
	{
		return;
	}

    if(cell->type == CELL_TYPE_ITEM)
    {
		gameScore += itemScore(cell->data.item);

        if(cell->data.item == ITEM_TYPE_OBJECTIVE_WATER_PIPE)
        {
            //update game state to indicate collected an objective item
            collectedObjectiveItems += 1;
        }
        else if(cell->data.item == ITEM_TYPE_OBJECTIVE_FUSE)
        {
            //update game state to indicate collected an objective item
			//note collectedObjectiveItems2 is separate to collectedObjectiveItems but only used for the last level
			//which has multiple primary objectives
            collectedObjectiveItems2 += 1;
        }
        else if(cell->data.item == ITEM_TYPE_PROJECTILE_LAUNCHER)
        {
            //enable projectile weapon
            player.hasProjectileLauncher = TRUE;

			//set current weapon to projectile launcher
			player.currentWeapon = PROJECTILE_WEAPON;

            //set initial ammo to 5
            player.projectileAmmo += 5;
        }
        else if(cell->data.item == ITEM_TYPE_PROJECTILE_AMMO)
        {
            //add 5 ammo (does not activate weapon pickup)
            player.projectileAmmo += 5;
        }
        else if(cell->data.item == ITEM_TYPE_FLAMETHROWER)
        {
            //enable flamethrower
            player.hasFlamethrower = TRUE;

			//set current weapon to flamethrower
			player.currentWeapon = FLAMETHROWER_WEAPON;

            //initial ammo is 5
            player.flamethrowerAmmo += 5;
        }
        else if(cell->data.item == ITEM_TYPE_FLAMETHROWER_AMMO)
        {
            //add 5 ammo (does not activate flamethrower weapon pickup)
            player.flamethrowerAmmo += 5;
        }
        else if(cell->data.item == ITEM_TYPE_LASER)
        {
			//enable laser
            player.hasLaser = TRUE;

			//set current weapon to laser
			player.currentWeapon = LASER_WEAPON;

            //initial ammo is 5
            player.laserAmmo += 5;
        }
        else if(cell->data.item == ITEM_TYPE_LASER_AMMO)
        {
            //add 5 ammo (does not activate laser weapon pickup)
            player.laserAmmo += 5;
        }
		else if (cell->data.item == ITEM_TYPE_KEY_RED)
		{
			player.inventory |= INVENTORY_ITEM_KEY;
		}
		else if (cell->data.item == ITEM_TYPE_KEY_GREEN)
		{
			player.inventory |= INVENTORY_ITEM_KEY_GREEN;
		}
		else if (cell->data.item == ITEM_TYPE_KEY_BLUE)
		{
			player.inventory |= INVENTORY_ITEM_KEY_BLUE;
		}
        else if (cell->data.item == ITEM_TYPE_MEDKIT)
        {
            if(player.health < PLAYER_MAX_HEALTH)
            {
                if(player.health + medkitHealth > PLAYER_MAX_HEALTH)
                {
                    player.health = startingHealth;
                }
                else
                {
                    player.health += medkitHealth;
                }
            }
        }
        else if (cell->data.item == ITEM_TYPE_OBJECTIVE_GEAR)
        {
            //gears
            //update game state to indicate collected an objective item
            collectedObjectiveItems += 1;
        }
        else if(cell->data.item == ITEM_TYPE_SPANNER)
        {
            //picked up a spanner and put it into the inventory
            player.inventory |= INVENTORY_ITEM_SPANNER;
        }
        else if(cell->data.item == ITEM_TYPE_OBJECTIVE_WIRE_SPOOL)
        {
            //wire spool
            //update game state to indicate collected an objective item
            collectedObjectiveItems += 1;
        }
		else if (cell->data.item == ITEM_TYPE_KEY_TEST_TUBE)
		{
			//picked up a test tube and put it into the inventory
			player.inventory |= INVENTORY_ITEM_TEST_TUBE;

			//turn off any flicker
			player.collisionIndicator = FALSE;

			//set game timer for time until test tube runs out
			player.timer = gameTimer + TEST_TUBE_TIME_;
		}

        //set cell to empty so it can be removed from the display and no longer processed
        cell->type = CELL_TYPE_EMPTY;
		
		//set the cell to empty tile for redraw
		cell->data.tile = emptyTile;

        //redraw cell without item
        drawCell(x, y, XY_TO_POSITION(x, y));

        //redraw hud with new score
        redrawHud = TRUE;
    }
	else if (cell->type == CELL_TYPE_INTERACTABLE)
	{
		if (cell->data.interactable == INTERACTABLE_TYPE_DOOR_RED)
		{
			//check player is carrying a key
			if (player.inventory & INVENTORY_ITEM_KEY)
			{
				//add some score
                gameScore += interactableScore(cell->data.interactable);

				//set cell to empty so it can be removed from the display and no longer processed
				cell->type = CELL_TYPE_EMPTY;

				//set the cell to empty tile for redraw
				cell->data.tile = emptyTile;

				//redraw cell without interactable
				drawCell(x, y, XY_TO_POSITION(x, y));

				//redraw hud with new score
				redrawHud = TRUE;
			}
		}
		else if (cell->data.interactable == INTERACTABLE_TYPE_DOOR_GREEN)
		{
			//check player is carrying a key
			if (player.inventory & INVENTORY_ITEM_KEY_GREEN)
			{
				//add some score
				gameScore += interactableScore(cell->data.interactable);

				//set cell to empty so it can be removed from the display and no longer processed
				cell->type = CELL_TYPE_EMPTY;

				//set the cell to empty tile for redraw
				cell->data.tile = emptyTile;

				//redraw cell without interactable
				drawCell(x, y, XY_TO_POSITION(x, y));

				//redraw hud with new score
				redrawHud = TRUE;
			}
		}
		else if (cell->data.interactable == INTERACTABLE_TYPE_DOOR_BLUE)
		{
			//check player is carrying a key
			if (player.inventory & INVENTORY_ITEM_KEY_BLUE)
			{
				//add some score
				gameScore += interactableScore(cell->data.interactable);

				//set cell to empty so it can be removed from the display and no longer processed
				cell->type = CELL_TYPE_EMPTY;

				//set the cell to empty tile for redraw
				cell->data.tile = emptyTile;

				//redraw cell without interactable
				drawCell(x, y, XY_TO_POSITION(x, y));

				//redraw hud with new score
				redrawHud = TRUE;
			}
		}
        else if (cell->data.interactable == INTERACTABLE_TYPE_DOOR_OBJECTIVE)
		{
			if (testGameObjectives())
			{
				//add some score
				gameScore += interactableScore(cell->data.interactable);

				nextLevel = TRUE;
			}			
            else
            {
				//ensures that the dialog is only displayed once
				if (objectiveDialogFlag_ == FALSE)
				{
					//display objective dialog
					objectiveDialog();

					//force a map redraw to cleanup after the dialog is closed
					mapRedraw = TRUE;

					//redraw the boss in case the objective dialog is displayed over the top of it
					gameBoss.redraw = TRUE;

					//flag set to true so dialog cannot be displayed until player moves onto exit again
					objectiveDialogFlag_ = TRUE;
				}                
            }
		}
        else if (cell->data.interactable == 2 || cell->data.interactable == INTERACTABLE_TYPE_GENERATOR_2_OFF || cell->data.interactable == INTERACTABLE_TYPE_GENERATOR_3_OFF)
        {
            //check the player has met the item objective 
			if (collectedObjectiveItems >= targetItemObjectiveCount)
			{
				//add some score
				gameScore += interactableScore(cell->data.interactable);

				//set the cell to the active version of the interactable
				cell->data.interactable += 1;

				//increase the activated tower interactable count
				++activatedTowerInteractableCount;

				//redraw cell with the new interactable sprite
				drawCell(x, y, XY_TO_POSITION(x, y));

				//redraw hud with new score
				redrawHud = TRUE;
			}
        }
        else if(cell->data.interactable == INTERACTABLE_TYPE_VALVE)
        {
            //add some score
			gameScore += interactableScore(cell->data.interactable);

            //set the cell to the active version of the interactable
            cell->data.interactable = INTERACTABLE_TYPE_VALVE_ACTIVE;

            //redraw cell with the new interactable sprite
            drawCell(x, y, XY_TO_POSITION(x, y));

			//update the game state following activation of the valve in the current segment
			++activatedValveCount;

			//removes the next lot of water from the map following valve activation
			removeWaterBlocker();

            //redraw hud with new score
            redrawHud = TRUE;
        }
        else if(cell->data.interactable == INTERACTABLE_TYPE_TRANSFORMER_0_OFF || cell->data.interactable == INTERACTABLE_TYPE_TRANSFORMER_1_OFF)
        {
			//check player is carrying a spanner
			if (player.inventory & INVENTORY_ITEM_SPANNER)
			{
				//check the player has met the item objective 
				if (collectedObjectiveItems >= targetItemObjectiveCount)
				{
					//add some score
					gameScore += interactableScore(cell->data.interactable);

					//set the cell to the active version of the interactable
					cell->data.interactable += 1;

					//increase the activated tower interactable count
					++activatedTowerInteractableCount;

					//redraw cell with the new interactable sprite
					drawCell(x, y, XY_TO_POSITION(x, y));

					//redraw hud with new score
					redrawHud = TRUE;
				}
			}
        }
		else if (cell->data.interactable == INTERACTABLE_TYPE_TESLA_COIL_0_OFF || cell->data.interactable == INTERACTABLE_TYPE_TESLA_COIL_1_OFF)
		{
			//check player is carrying a spanner
			if (player.inventory & INVENTORY_ITEM_SPANNER)
			{
				//check the player has met the item objective 
				if (collectedObjectiveItems2 >= targetItemObjectiveCount)
				{
					//add some score
					gameScore += interactableScore(cell->data.interactable);

					//set the cell to the active version of the interactable
					cell->data.interactable += 1;

					//increase the activated tower interactable count
					++activatedTowerInteractableCount;

					//redraw cell with the new interactable sprite
					drawCell(x, y, XY_TO_POSITION(x, y));

					//redraw hud with new score
					redrawHud = TRUE;
				}
			}
		}
		else if (cell->data.interactable == INTERACTABLE_TYPE_WATER_TANK_0_EMPTY)
		{
			//check player is carrying a spanner
			if (player.inventory & INVENTORY_ITEM_SPANNER)
			{
				//check the player has met the item objective 
				if (collectedObjectiveItems >= targetItemObjectiveCount)
				{
					//add some score
					gameScore += interactableScore(cell->data.interactable);

					//set the cell to the active version of the interactable
					cell->data.interactable += 1;

					//increase the activated tower interactable count
					++activatedTowerInteractableCount;

					//redraw cell with the new interactable sprite
					drawCell(x, y, XY_TO_POSITION(x, y));

					//redraw hud with new score
					redrawHud = TRUE;
				}
			}
		}
		else if (cell->data.interactable == INTERACTABLE_TYPE_DEATH)
		{
			if (player.health > 2)
			{
				//lose 2 health
				player.health -= 2;
			}
			else
			{
				player.health = 0;
			}			

			//redraw hud to update with new health value
			redrawHud = TRUE;
		}
	}
	else if (cell->type == CELL_TYPE_ENEMY)
	{
		if(testObjectIntersection(&player.object, &enemies[cell->data.enemy].object))
		{
			//note there seems to be a potential for a 1 pixel gap displayed when they "colide" but this seems like a compromise im willing to have
#ifdef _WIN32
			//printf("enemy collision!!!!!!!!!!!!!!!!!!!!!\n");
#endif
			//set enemy as stunned from the collision
			enemies[cell->data.enemy].stunned = TRUE;

			//timer is set to 1000 increments of the timer into the future
			enemies[cell->data.enemy].timer = gameTimer + ENEMY_STUN_TIME;

			//while the player is carrying a test tube they do not take damage
			if (player.inventory & INVENTORY_ITEM_TEST_TUBE)
			{
				return 2;
			}

			//playerCollisionFlickerCount effectively serves as a player invincibility timer
			if (player.collisionFlickerCount == 0)
			{
				player.collisionIndicator = TRUE;
				player.timer = gameTimer + 50;

				//number of times to flicker between the player graphic and white
				player.collisionFlickerCount = 5;

				//subtract 1 from player health
				if (player.health)
				{
					--player.health;

					//redraw hud to update with new health value
					redrawHud = TRUE;
				}
			}

			//found an enemy
			return 2;
		}		

		//no enemy found
		return 1;
	}
	else if (cell->type == CELL_TYPE_BOSS)
	{
		//player has collided with the boss
		return BOSS_CELL_FOUND_;
	}

	//not enemy cell
	return 0;
}

void playerShoot()
{
	switch (player.currentWeapon)
	{
		case PROJECTILE_WEAPON:
		{
			if (player.projectileAmmo)
			{
				player.projectile.projectileActive = TRUE;
			}
		}
		break;

		case FLAMETHROWER_WEAPON:
		{
			if (player.flamethrowerAmmo)
			{
				//make sure the flame isn't already displaying
				if (player.projectile.type == 0)
				{
					//shoot the flamethrower and lower the flamethrower ammo count
					--player.flamethrowerAmmo;

					initProjectile(&player.projectile, FLAME_PROJECTILE, player.object.position.x, player.object.position.y, player.direction, player.object.width, player.object.height, flameBackgroundMemory_);

					//redraw the hud to show the updated ammo value
					redrawHud = TRUE;
				}
			}
		}
		break;

		case LASER_WEAPON:
		{
			if (player.laserAmmo)
			{
                //make sure the laser isnt already displaying
                if(player.projectile.type == 0)
                {
                    //shoot the laser and lower the lasser ammo count
                    --player.laserAmmo;

                    initProjectile(&player.projectile, LASER_PROJECTILE, player.object.position.x, player.object.position.y, player.direction, player.object.width, player.object.height, laserBackgroundMemory_);

                    //redraw the hud to show the updated ammo value
                    redrawHud = TRUE;
                }
			}
		}			
		break;
    }
}

void playerEndShoot()
{
    if(player.currentWeapon == LASER_WEAPON)
    {
        //make sure there is a projectile
        if(player.projectile.type == LASER_PROJECTILE)
        {
            //redraw the background
			drawProjectileBackgroundMemoryAtPosition(&player.projectile);

			//end the current projectile
			player.projectile.type = 0;
        }
    }
    else if(player.currentWeapon == FLAMETHROWER_WEAPON)
    {
		//make sure there is a projectile
		if (player.projectile.type == FLAME_PROJECTILE)
		{
			if (player.projectile.infestationFlag == FALSE)
			{
				//redraw the background
				drawProjectileBackgroundMemoryAtPosition(&player.projectile);
			}
			else
			{
				//draw any cells currently queued
				dequeueDrawQueue();

				//updates the background
				copyProjectileBackground(&player.projectile);

				//redraw the background
				drawProjectileBackgroundMemoryAtPosition(&player.projectile);

				//force hud redraw incase the flame projectile is drawn on top of hud
				redrawHud = TRUE;
			}

			//end the current projectile
			player.projectile.type = 0;
		}
    }
	else if (player.currentWeapon == PROJECTILE_WEAPON)
	{
		if (player.projectile.projectileActive)
		{
			//todo determine if we are happy with only one projectile on the screen at once.
			//this particular line ensures that only one projectile can be on screen at any one time.
			if (player.projectile.type == 0)
			{
				--player.projectileAmmo;

				initProjectile(&player.projectile, BULLET_PROJECTILE, player.object.position.x, player.object.position.y, player.direction, player.object.width, player.object.height, bulletBackgroundMemory);

				//redraw the hud to show the updated ammo value
				redrawHud = TRUE;
			}

			player.projectile.projectileActive = FALSE;
		}
	}
}

void determinePlayerPosition(u8* x, u8* y)
{
	u8 i = 0;
	u8 j = 0;

	//find the first empty space on the grid
	for (i = 0; i < SCREEN_TILES_X; ++i)
	{
		for (j = 0; j < SCREEN_TILES_Y; ++j)
		{
			if (mapCells[XY_TO_POSITION(i, j)].type == CELL_TYPE_EMPTY)
			{
				*x = i;
				*y = j;

				return;
			}
		}
	}
}

void weaponSwitch()
{
	if (gameControls.weaponIsDown)
	{
		if (gameTimer >= weaponSwitchTimer)
		{
			//switches which weapon the player is currently using based on what they are carrying
			//not checking ammo, that is the player's problem
			if (player.currentWeapon == PROJECTILE_WEAPON)
			{
				if (player.hasFlamethrower)
				{
					player.currentWeapon = FLAMETHROWER_WEAPON;
				}
				else if (player.hasLaser)
				{
					player.currentWeapon = LASER_WEAPON;
				}
			}
			else if (player.currentWeapon == FLAMETHROWER_WEAPON)
			{
				if (player.hasLaser)
				{
					player.currentWeapon = LASER_WEAPON;
				}
				else if (player.hasProjectileLauncher)
				{
					player.currentWeapon = PROJECTILE_WEAPON;
				}
			}
			else if (player.currentWeapon == LASER_WEAPON)
			{
				if (player.hasProjectileLauncher)
				{
					player.currentWeapon = PROJECTILE_WEAPON;
				}
				else if (player.hasFlamethrower)
				{
					player.currentWeapon = FLAMETHROWER_WEAPON;
				}
			}

			//force a hud redraw following weapon switch
			redrawHud = TRUE;

			weaponSwitchTimer = gameTimer + WEAPON_SWITCH_TIME_;
		}
	}
}