#include "Projectile.h"
#include <stdio.h>
#include <string.h>

void initProjectile(Projectile* projectile, u8 type, u8 x, u8 y, u8 direction, u8 playerWidth, u8 playerHeight, u8* graphicsMemory)
{
	projectile->type = type;
	projectile->direction = direction;
	projectile->infestationFlag = FALSE;

	if (type == BULLET_PROJECTILE)
	{
		projectile->object.width = BULLET_PROJECTILE_WIDTH;
		projectile->object.height = BULLET_PROJECTILE_HEIGHT;
	}
	else if (type == LASER_PROJECTILE)
	{
		//start as the same size as the bullet for now
		projectile->object.width = BULLET_PROJECTILE_WIDTH;
		projectile->object.height = BULLET_PROJECTILE_HEIGHT;
	}
    
	//we assume that the graphics memory is initialised outside of the function
	projectile->backgroundMemory = graphicsMemory;


	if (direction == DIRECTION_SOUTH)
	{
		projectile->object.velocity.x = 0;
		projectile->object.velocity.y = 2;
		projectile->object.position.x = x;
		projectile->object.position.y = y + playerHeight + 1;
	}
	else if (direction == DIRECTION_NORTH)
	{
		projectile->object.velocity.x = 0;
		projectile->object.velocity.y = -2;
		projectile->object.position.x = x;
		projectile->object.position.y = y - 1;
	}
	else if (direction == DIRECTION_EAST)
	{
		projectile->object.velocity.x = 1;
		projectile->object.velocity.y = 0;
		projectile->object.position.x = x + playerWidth + 1;
		projectile->object.position.y = y;
	}
	else
	{
		//west
		projectile->object.velocity.x = -1;
		projectile->object.velocity.y = 0;
		projectile->object.position.x = x - 1;
		projectile->object.position.y = y;
	}

	projectile->object.previousPosition.x = projectile->object.position.x;
	projectile->object.previousPosition.y = projectile->object.position.y;


	if (projectile->type == LASER_PROJECTILE)
	{
		//for calculating the max size of laser projfile
		u8 cellX = 0;
		u8 cellY = 0;
		i8 testInterval = 0;

		u8 tempX = projectile->object.position.x;
		u8 tempY = projectile->object.position.y;

		if (projectile->direction == DIRECTION_NORTH || projectile->direction == DIRECTION_WEST)
		{
			testInterval = -1;
		}
		else if (projectile->direction == DIRECTION_SOUTH || projectile->direction == DIRECTION_EAST)
		{
			testInterval = 1;
		}

		do
		{
			if (projectile->direction == DIRECTION_NORTH || projectile->direction == DIRECTION_SOUTH)
			{
				tempY += testInterval;
			}

			if (projectile->direction == DIRECTION_EAST || projectile->direction == DIRECTION_WEST)
			{
				tempX += testInterval;
			}

			cellX = tempX / TILE_WIDTH_BYTES;
			cellY = tempY / TILE_HEIGHT;
		} 
		while (MAPA(cellX, cellY).type == CELL_TYPE_EMPTY);

		if (MAPA(cellX, cellY).type == CELL_TYPE_BOSS)
		{
			gameBoss.health -= 5;
		}

		if (projectile->direction == DIRECTION_NORTH)
		{
			projectile->object.height = projectile->object.position.y - tempY;
			
			//update position.y with max y value (tempY - testInterval)
			projectile->object.position.y = tempY - testInterval;
		}
		else if (projectile->direction == DIRECTION_SOUTH)
		{
			projectile->object.height = tempY - projectile->object.position.y;
		}
		else if (projectile->direction == DIRECTION_EAST)
		{
			projectile->object.width = tempX - projectile->object.position.x;
		}
		else
		{
			//west
			projectile->object.width = projectile->object.position.x - tempX;

			//update position.x with max x value (tempx - testInterval)
			projectile->object.position.x = tempX - testInterval;
		}
	}
	else if (projectile->type == FLAME_PROJECTILE)
	{
		if (projectile->direction == DIRECTION_NORTH || projectile->direction == DIRECTION_SOUTH)
		{
			projectile->object.width = FLAME_PRJECTILE_VERTICAL_WIDTH;
			projectile->object.height = FLAME_PRJECTILE_VERTICAL_HEIGHT;

			if (projectile->direction == DIRECTION_NORTH)
			{
				//make sure flame can't go off the screen (up)
				if (projectile->object.position.y - FLAME_PRJECTILE_VERTICAL_HEIGHT > 0)
				{
					projectile->object.position.y -= FLAME_PRJECTILE_VERTICAL_HEIGHT;
				}
				else
				{
					projectile->object.position.y = 0;
				}				
			}

			//make sure flame can't go off the screen (bottom)
			if (projectile->object.position.y + projectile->object.height > SCREEN_Y_NO_HUD)
			{
				projectile->object.height = SCREEN_Y_NO_HUD - projectile->object.position.y;
			}
		}
		else
		{
			//east or west
			projectile->object.width = FLAME_PRJECTILE_HORIZONTAL_WIDTH;
			projectile->object.height = FLAME_PRJECTILE_HORIZONTAL_HEIGHT;

			if (projectile->direction == DIRECTION_WEST)
			{
				//make sure flame can't go off the screen (left)
				if (projectile->object.position.x - FLAME_PRJECTILE_HORIZONTAL_WIDTH > 0)
				{
					projectile->object.position.x -= FLAME_PRJECTILE_HORIZONTAL_WIDTH;
				}
				else
				{
					projectile->object.position.x = 0;
				}
			}

			//make sure flame can't go off the screen (right)
			if (projectile->object.position.x + projectile->object.width > SCREEN_X)
			{
				projectile->object.width = SCREEN_X - projectile->object.position.x;
			}
		}
	}

	projectile->projectileActive = FALSE;

	//copy the background at the projectiles position
	copyProjectileBackground(projectile);
}

void copyProjectileBackground(Projectile* projectile)
{
	u8 i = 0;

	//copy the content of the background where the projectile currently is located
	for (i = 0; i < projectile->object.height; ++i)
	{
		cpct_memcpy(projectile->backgroundMemory + (i * projectile->object.width), cpct_getScreenPtr(CPCT_VMEM_START, projectile->object.position.x, projectile->object.position.y + i), projectile->object.width);
	}
}

void copyPreviousProjectileBackground(Projectile* projectile)
{
	u8 i = 0;

	//copy the content of the background where the projectile currently is located
	for (i = 0; i < projectile->object.height; ++i)
	{
		cpct_memcpy(projectile->backgroundMemory + (i * projectile->object.width), cpct_getScreenPtr(CPCT_VMEM_START, projectile->object.previousPosition.x, projectile->object.previousPosition.y + i), projectile->object.width);
	}
}

void drawProjectile(Projectile* projectile)
{
	//projectile shouldn't be drawn if it has been "destroyed"
	if (projectile->type != 0)
	{
		u8* memory = cpct_getScreenPtr(CPCT_VMEM_START, projectile->object.position.x, projectile->object.position.y);

		if (projectile->type == BULLET_PROJECTILE)
		{
			//draw a square to represent the projectile		
			cpct_drawSolidBox(memory, cpct_px2byteM0(3, 3), projectile->object.width, projectile->object.height);
		}
		else if (projectile->type == LASER_PROJECTILE)
		{
			//draw a rectangle to represent the laser beam			
			cpct_drawSolidBox(memory, cpct_px2byteM0(3, 3), projectile->object.width, projectile->object.height);
		}
		else if (projectile->type == FLAME_PROJECTILE)
		{
			//draw a blue rectangle to represent the flame
			cpct_drawSolidBox(memory, cpct_px2byteM0(2, 2), projectile->object.width, projectile->object.height);
		}
	}
}

void drawPreviousBackgroundMemory(Projectile* projectile)
{
	//only draw the projectile if it actually exists
	if(projectile->type != 0)
	{
		u8* memory = cpct_getScreenPtr(CPCT_VMEM_START, projectile->object.previousPosition.x, projectile->object.previousPosition.y);
		cpct_drawSprite(projectile->backgroundMemory, memory, projectile->object.width, projectile->object.height);
	}	
}

void drawProjectileBackgroundMemoryAtPosition(Projectile* projectile)
{
	//only draw the projectile if it actually exists
	if (projectile->type != 0)
	{
		u8* memory = cpct_getScreenPtr(CPCT_VMEM_START, projectile->object.position.x, projectile->object.position.y);
		cpct_drawSprite(projectile->backgroundMemory, memory, projectile->object.width, projectile->object.height);
	}
}

void logicProjectile(Projectile* projectile, Object* parent)
{    
	//logic only needs to be run if the projectile actually moves down the screen
	if(projectile->type == BULLET_PROJECTILE)
	{
        u8 cellX = 0;
        u8 cellY = 0;

		//set previous position for the projectile
		projectile->object.previousPosition.x = projectile->object.position.x;
		projectile->object.previousPosition.y = projectile->object.position.y;

		//calculate new projectile position from velocity		
        projectile->object.position.y += projectile->object.velocity.y;
        projectile->object.position.x += projectile->object.velocity.x;

		if (projectile->object.position.y < 0 || projectile->object.position.y > SCREEN_Y_NO_HUD)
		{
			//y is now at end of range so destroy the projectile
			destroyBulletProjectile(projectile);
			return;
		}

		if (projectile->object.position.x < 0 || projectile->object.position.x > SCREEN_X)
		{
			//x is now at end of range so destroy the projectile
			destroyBulletProjectile(projectile);
			return;
		}

        //check whether the new position is in a solid cell or not
        //if its solid we change the state to inactive

		cellX = projectile->object.position.x / TILE_WIDTH_BYTES;
		cellY = projectile->object.position.y / TILE_HEIGHT;

		if (MAPA(cellX, cellY).type == CELL_TYPE_ENEMY)
		{
			Enemy* enemy = &enemies[MAPA(cellX, cellY).data.enemy];
			
			if (testObjectIntersection(&enemy->object, &projectile->object))
			{
				//reached an enemy so the projectile should be destroyed/set to inactive
				destroyBulletProjectile(projectile);

				//kill enemy and cleanup
				killEnemy(enemy);

				//remove enemy from current segment
				--segmentEnemies[currentSegment];
			}
		}
		else if (MAPA(cellX, cellY).type == CELL_TYPE_BOSS)
		{
			//reduce boss health on projectile hit
			--gameBoss.health;

			//reached the boss so the projectile should be destroyed/set to inactive
			destroyBulletProjectile(projectile);
		}
		else if (MAPA(cellX, cellY).type != CELL_TYPE_EMPTY)
        {
			//reached an obstacle so the projectile should be destroyed/set to inactive
			destroyBulletProjectile(projectile);
        }		
	}
	else if (projectile->type == FLAME_PROJECTILE)
	{	
		u8 cellXMin = 0;
		u8 cellXMax = 0;
		u8 cellYMin = 0;
		u8 cellYMax = 0;
		u8 i = 0;
		u8 j = 0;

		//set previous x position for the projectile
		projectile->object.previousPosition.x = projectile->object.position.x;
		projectile->object.previousPosition.y = projectile->object.position.y;
		
		//vertical projectile
		if (projectile->direction == DIRECTION_NORTH || projectile->direction == DIRECTION_SOUTH)
		{
			//make sure any change of position won't go off the screen
			if (projectile->object.position.x + projectile->object.velocity.x > 0 && projectile->object.position.x + projectile->object.velocity.x < SCREEN_X)
			{
				//the flame is only allowed to go one width to the left or one width to the right
				if (projectile->object.position.x + projectile->object.velocity.x >= parent->position.x - FLAME_PRJECTILE_VERTICAL_WIDTH &&
					projectile->object.position.x + projectile->object.velocity.x <= parent->position.x + parent->width)
				{
					projectile->object.position.x += projectile->object.velocity.x;
				}
			}
		}
		else
		{
			//horizontal projectile

			//make sure any change of position won't go off the screen
			if (projectile->object.position.y + projectile->object.velocity.y > 0 && projectile->object.position.y + projectile->object.velocity.y < SCREEN_Y_NO_HUD)
			{
				//the flame is only allowed to go one height to the left or one height to the right
				if (projectile->object.position.y + projectile->object.velocity.y >= parent->position.y - 12 &&
					projectile->object.position.y + projectile->object.velocity.y <= parent->position.y + 16)
				{
					projectile->object.position.y += projectile->object.velocity.y;
				}
			}
		}

		cellXMin = projectile->object.position.x >> 2;
		cellXMax = (projectile->object.position.x + projectile->object.width);
		cellYMin = projectile->object.position.y >> 4;
		cellYMax = (projectile->object.position.y + projectile->object.height);
		cellXMax >>= 2;
		cellYMax >>= 4;

		for (i = cellXMin; i <= cellXMax; ++i)
		{
			for (j = cellYMin; j <= cellYMax; ++j)
			{
				if (MAPA(i, j).type == CELL_TYPE_TILE)
				{
					if (MAPA(i, j).data.tile == TILE_TYPE_WALL_MIDDLE_INFESTED)
					{
						MAPA(i, j).data.tile = TILE_TYPE_BURNT_WALL;

						projectile->infestationFlag = TRUE;
 						--infestedCellCount;
					}
					else if (MAPA(i, j).data.tile == TILE_TYPE_EMPTY_MIDDLE_INFESTED)
					{
						MAPA(i, j).type = CELL_TYPE_EMPTY;

						projectile->infestationFlag = TRUE;
						--infestedCellCount;
					}
				}
				else if (MAPA(i, j).type == CELL_TYPE_ENEMY)
				{
					Enemy* enemy = &enemies[MAPA(i, j).data.enemy];

					if (testObjectIntersection(&enemy->object, &projectile->object))
					{
						enemy->stunned = TRUE;
						enemy->timer = gameTimer + ENEMY_STUN_TIME;
					}
				}

				//queue it all just in case
				addCellToDrawQueue(i, j);
			}
		}
		
		//reset velocity following change of position
		projectile->object.velocity.x = 0;
		projectile->object.velocity.y = 0;
	}
}

void destroyBulletProjectile(Projectile* projectile)
{
	u8* memory;

	//reached an obstacle so the projectile should be destroyed/set to inactive
	projectile->type = 0;

	//redraw the background where the sprite is currently to make it disappear
	//minus the velocity to get us the previous position
	memory = cpct_getScreenPtr(CPCT_VMEM_START, projectile->object.position.x - projectile->object.velocity.x, projectile->object.position.y - projectile->object.velocity.y);
	cpct_drawSprite(projectile->backgroundMemory, memory, projectile->object.width, projectile->object.height);
}